;; avg_trace_lengths.scm
;;
;; This file is an CLI executable command.  
;;
;; It lists the average number of text indices accessed divided by the
;; length of the text applied with

(use-modules (ice-9 pretty-print))
(use-modules (srfi srfi-37))

(load (string-append (getcwd) "/framework/init.scm"))

(define (p-stderr . args)
  (if (null? args)
      #t
      (let ((stderr (current-error-port)))
        (write (car args) stderr) 
        (display " " stderr) 
        (apply p (cdr args)))))

(define (pn-stderr . args)
  (apply p-stderr args)
  (newline (current-error-port)))

(define pat-abc "ab")
(define pat-length 4)
(define txt-abc "abc")
(define txt-length 5)

(define (load-matcher file)
  (let ((matchers (load (if (eq? #\/ (string-ref file 0))
                           file
                           (string-append (dirname file) "/../" 
                                          (basename file))))))
    (verify-matchers matchers)
    matchers))

(define (verify-matchers matchers)
  (let ((err (lambda (m error)
               (error "Error in matchers entry '" m "': " error))))
    (for-each (lambda (m) 
                (cond
                 ((not (and (= (length m) 2) 
                            (string? (car m))
                            (procedure? (cadr m))))
                  (err m (string-append "Needs to be of the form "
                                        "(\"matcher-name\" procedure).")))
                 ((not (list? ((cadr m) "a" "a")))
                  (err m (string-append "Procedure returned non-list output, "
                                        "needs to return a lst of integers.")))
                 (else
                  #t)))
              matchers)))

(define (get-opts-traces assoc-or actual-args)
  (let* ((matchers (load-matcher (car actual-args)))
         (patabc (assoc-or "patabc" pat-abc))
         (patlength (assoc-or "patlength" pat-length))
         (txtabc (assoc-or "txtabc" txt-abc))
         (txtlength (assoc-or "txtlength" txt-length))
         (lazy-pats (string-permutations-lazy patabc 3 patlength))
         (lazy-txts (string-permutations-lazy txtabc 1 txtlength))
         (output (assoc-or "output" "text")))
    (cond
      ((string= output "text")
       (output-text (avg-trace-lengths matchers lazy-pats lazy-txts)))
      (else
       (error "Unknown output method")))))

(define (output-text trace-lengths)
  (fold (lambda (tl s) (format "~a~a ~f\n" s (car tl) (cdr tl)))
        "" trace-lengths))

(define (make-dmatrix distance-proc matchers lazy-pats lazy-txts)
  (dmatrix->phylip matchers 
                   (distance-matrix distance-proc matchers 
                                    lazy-pats lazy-txts)))

(define (build-tree matchers dmatrix)
  (my-display-file temp-file dmatrix)
  (rename-tree matchers
               (run-external quicktree-path 
                             (string-append "-in m -out t " temp-file))))

(define (main args)
  (let* ((name-cons (lambda (name)
                      (lambda (opt name' arg loads) 
                        (cons (cons name arg) loads))))
         (name-cons-int (lambda (name)
                      (lambda (opt name' arg loads) 
                        (cons (cons name (string->number arg)) loads))))
         (opts (args-fold 
                (cdr args)
                (list 
                 (option '(#\h "help") #f #f (name-cons "help"))
                 (option '(#\a "patabc") #t #f (name-cons "patabc"))
                 (option '(#\l "patlength") #t #f (name-cons-int "patlength"))
                 (option '(#\A "txtabc") #t #f (name-cons "txtabc"))
                 (option '(#\L "txtlength") #t #f (name-cons-int "txtlength"))
                 (option '(#\o "output") #t #f (name-cons "output"))
                 )
                (lambda (opt name arg loads)
                  (error (format #f "Unrecognized option `~A'" name)))
                (lambda (arg loads) (cons (cons "arg" arg) loads))
                '()))
         (actual-args (reverse (map cdr (filter (lambda (opt) 
                                                  (equal? (car opt) "arg"))
                                                opts))))
         (number-of-args (length actual-args)))
    (cond
     ((or (assoc "help" opts) (= 0 number-of-args))
      (display-help))
     (else
      (let* ((assoc-or (lambda (name or-val)
                         (let ((opt (assoc name opts)))
                           (cond
                            ((not opt) or-val)
                            ((not (cdr opt)) #t)
                            (else (cdr opt)))))))
        (display (get-opts-traces assoc-or actual-args)))))))

(define (display-help)
  (display "Usage: program [OPTIONS] MATCHERS-FILE
Generates an table showing average trace-length / text length for
matchers in MATCHERS-FILE.

  -h, --help                print this message

  -a, --patabc=ALPHABET     the pattern alphabet, default is ab
  -l, --patlength=SIZE      the pattern lengths, from 3 to SIZE, default is 3
  -A, --txtabc=ALPHABET     the text alphabet, default is abc
  -L, --txtlength=SIZE      the text length, default is 4

  -o, --output=OUTPUT       default: text, options: text

EXAMPLE: Default usage
    ./avg_trace_lengths matchers.scm
"))

    
(main (command-line))
