;; identify_matcher.scm
;;
;; This file is an CLI executable command.  
;;
;; It compares a single matcher against a set of other matchers over a
;; given set of inputs.

(use-modules (ice-9 pretty-print))
(use-modules (srfi srfi-37))

(load (string-append (getcwd) "/framework/init.scm"))

(define (p-stderr . args)
  (if (null? args)
      #t
      (let ((stderr (current-error-port)))
        (write (car args) stderr) 
        (display " " stderr) 
        (apply p (cdr args)))))

(define (pn-stderr . args)
  (apply p-stderr args)
  (newline (current-error-port)))

(define pat-abc "ab")
(define pat-length 4)
(define txt-abc "abc")
(define txt-length 5)
(define default-gapcost 2)
(define default-diffcost 5)

(define quicktree-path "quicktree/bin/quicktree")
(define temp-file "/tmp/evolutionary_tree.scm.tmp")

(define (load-matcher file)
  (let ((matchers (load (if (eq? #\/ (string-ref file 0))
                           file
                           (string-append (dirname file) "/../" 
                                          (basename file))))))
    (verify-matchers matchers)
    matchers))

(define (verify-matchers matchers)
  (let ((err (lambda (m error)
               (error "Error in matchers entry '" m "': " error))))
    (for-each (lambda (m) 
                (cond
                 ((not (and (= (length m) 2) 
                            (string? (car m))
                            (procedure? (cadr m))))
                  (err m (string-append "Needs to be of the form "
                                        "(\"matcher-name\" procedure).")))
                 ((not (list? ((cadr m) "a" "a")))
                  (err m (string-append "Procedure returned non-list output, "
                                        "needs to return a lst of integers.")))
                 (else
                  #t)))
              matchers)))

(define (get-opts-traces assoc-or actual-args)
  (let* ((matchers (load-matcher (car actual-args)))
         (newick-file (assoc-or "newick" #f))
         (newick (and newick-file (my-read-file-as-string newick-file)))
         (patabc (assoc-or "patabc" pat-abc))
         (patlength (assoc-or "patlength" pat-length))
         (txtabc (assoc-or "txtabc" txt-abc))
         (txtlength (assoc-or "txtlength" txt-length))
         (lazy-pats (string-permutations-lazy patabc 3 patlength))
         (lazy-txts (string-permutations-lazy txtabc 1 txtlength))
         (output (assoc-or "output" "dot"))
         (distance (assoc-or "distance" "naive"))
         (gapcost (assoc-or "gapcost" default-gapcost))
         (diffcost (assoc-or "diffcost" default-diffcost))
         (distance-proc (cond
                         ((string= distance "naive") 
                          trace-update-equal-or-not)
                         ((string= distance "alignment") 
                          (trace-update-pairwise-alignment gapcost diffcost))
                         (else (error "Unknown distance method")))))
    (cond
      ((string= output "dmatrix")
       (make-dmatrix distance-proc matchers lazy-pats lazy-txts))
      ((string= output "newick2dot")
       (dot-parsed-tree
        (parse-tree (rename-tree matchers newick))))
      ((string= output "dot")
       (dot-parsed-tree
        (parse-tree
         (build-tree matchers 
                     (make-dmatrix distance-proc matchers 
                                   lazy-pats lazy-txts)))))
      (else
       (error "Unknown output method")))))

(define (make-dmatrix distance-proc matchers lazy-pats lazy-txts)
  (dmatrix->phylip matchers 
                   (distance-matrix distance-proc matchers 
                                    lazy-pats lazy-txts)))

(define (build-tree matchers dmatrix)
  (my-display-file temp-file dmatrix)
  (rename-tree matchers
               (run-external quicktree-path 
                             (string-append "-in m -out t " temp-file))))

(define (main args)
  (let* ((name-cons (lambda (name)
                      (lambda (opt name' arg loads) 
                        (cons (cons name arg) loads))))
         (name-cons-int (lambda (name)
                      (lambda (opt name' arg loads) 
                        (cons (cons name (string->number arg)) loads))))
         (opts (args-fold 
                (cdr args)
                (list 
                 (option '(#\h "help") #f #f (name-cons "help"))
                 (option '(#\a "patabc") #t #f (name-cons "patabc"))
                 (option '(#\l "patlength") #t #f (name-cons-int "patlength"))
                 (option '(#\A "txtabc") #t #f (name-cons "txtabc"))
                 (option '(#\L "txtlength") #t #f (name-cons-int "txtlength"))
                 (option '(#\o "output") #t #f (name-cons "output"))
                 (option '(#\n "newick") #t #f (name-cons "newick"))
                 (option '(#\d "distance") #t #f (name-cons "distance"))
                 (option '(#\D "diffcost") #t #f (name-cons-int "diffcost"))
                 (option '(#\g "gapcost") #t #f (name-cons-int "gapcost"))
                 )
                (lambda (opt name arg loads)
                  (error (format #f "Unrecognized option `~A'" name)))
                (lambda (arg loads) (cons (cons "arg" arg) loads))
                '()))
         (actual-args (reverse (map cdr (filter (lambda (opt) 
                                                  (equal? (car opt) "arg"))
                                                opts))))
         (number-of-args (length actual-args)))
    (cond
     ((or (assoc "help" opts) (= 0 number-of-args))
      (display-help))
     (else
      (let* ((assoc-or (lambda (name or-val)
                         (let ((opt (assoc name opts)))
                           (cond
                            ((not opt) or-val)
                            ((not (cdr opt)) #t)
                            (else (cdr opt)))))))
        (display (get-opts-traces assoc-or actual-args)))))))

(define (display-help)
  (display "Usage: program [OPTIONS] MATCHERS-FILE
Generates an evolutionary tree over the matchers in MATCHERS-FILE.

  -h, --help                print this message

  -a, --patabc=ALPHABET     the pattern alphabet, default is ab
  -l, --patlength=SIZE      the pattern lengths, from 3 to SIZE, default is 3
  -A, --txtabc=ALPHABET     the text alphabet, default is abc
  -L, --txtlength=SIZE      the text length, default is 4

  -o, --output=OUTPUT       default: dot, options: dot, dmatrix, rename
  -d, --distance=METHOD     default: naive, options: naive, alignment

Extra options for the alignment distance method:

  -D, --diffcost=COST       cost of different characters, default: 5
  -g, --gapcost=COST        cost of removing/inserting characters, default: 2

EXAMPLE: Make an evolutionary tree
    ./evolutionary_tree matchers.scm
"))

    
(main (command-line))
