;; framework_comparison.scm
;; 
;; This file defines functions used to compare string matchers.
;;
;; This file defines implement these concepts:
;; - lazily generating string permutatinos
;; - generate table separating string matchers
;; - identify a single string matcher
;; - generate distance-matrix

(use-modules (ice-9 format))

;;;; Generate string permutations
(define lazy-car car)
(define (lazy-cdr e)
  ((cadr e)))
(define (lazy-null? e)
  (null? (car e)))
(define (lazy->list e)
  (if (lazy-null? e)
      '()
      (cons (lazy-car e) (lazy->list (lazy-cdr e)))))
(define (list->lazy l)
  (if (null? l)
      (list '() '())
      (list (car l) (lambda () (list->lazy (cdr l))))))

(define (string-permutations-lazy abc length-from length-to)
  (letrec ((walk
            (lambda (abc0 i str f)
              (cond
               ((null? abc0)
                (f))
               ((= i 0)
                (list str f))
               (else
                (walk abc (- i 1) (string-append str (car abc0))
                      (lambda ()
                        (walk (cdr abc0) i str
                              (lambda ()
                                (f)))))))))
           (walk1 (lambda (length f)
                    (if (> length length-to)
                        (list '() #f)
                        (walk abc length ""
                              (lambda ()
                                (walk1 (1+ length) f)))))))
    (if (string? abc)
        (string-permutations-lazy
         (map char->string (string->list abc)) length-from length-to)
        (walk1 length-from (lambda () (list '() #f))))))

;;;; generate traces for a gives set of matchers, a pattern and a text
(define (list<-maker f< f=)
  (letrec ((walk (lambda (a b)
                   (cond ((and (not (pair? a)) (not (pair? b))) #f)
                         ((and (not (pair? a)) (pair? b)) #t)
                         ((and (pair? a) (not (pair? b))) #f)
                         ((f= (car a) (car b)) (walk (cdr a) (cdr b)))
                         (else (f< (car a) (car b)))))))
    walk))

(define (group-same-trace xs)
  (letrec ((walk (lambda (xs last-trace out)
                   (if (null? xs)
                       (cons (cons last-trace (car out))
                             (cdr out))
                       (let ((name (caar xs))
                             (trace (cdar xs)))
                         (if (equal? trace last-trace)
                             (walk (cdr xs) trace
                                   (cons (cons name (car out))
                                         (cdr out)))
                             (if (null? out)
                                 (walk (cdr xs) trace
                                       (list (list name)))
                                 (walk (cdr xs) trace
                                       (cons
                                        (list name)
                                        (cons
                                         (cons last-trace (car out))
                                         (cdr out)))))))))))
    (let ((trace< (list<-maker < =)))
      (walk (sort xs (lambda (a b) (trace< (cdr a) (cdr b))))
            #f '()))))

(define (compare-matchers matchers pattern text)
  (let ((grouped-matchers (group-same-trace
                           (map (lambda (matcher)
                                  (let ((name (car matcher))
                                        (proc (cadr matcher)))
                                    (cons name (proc pattern text))))
                                matchers))))
    (list (list pattern text)
          grouped-matchers)))

;;;; Trim traces that are doesn't add information (to limit memory use)
(define (prune-traces traces)
  (let ((used-inputs (map (lambda (t) (list-head (caadr t) 2))
                          (pretty-traces traces))))
    (filter (lambda (t)
              (find (lambda (t')
                      (equal? t' (car t)))
                    used-inputs))
            traces)))

(define (sort-traces traces)
  (let ((sorter (lambda (a b)
                   (let* ((groups (lambda (t) (length (cadr t))))
                          (inputsize (lambda (t) (+ (string-length (caar t))
                                                    (string-length (cadar t)))))
                          (groups-a (groups a))
                          (groups-b (groups b))
                          (inputsize-a (inputsize a))
                          (inputsize-b (inputsize b))
                          (pat-a (caar a))
                          (txt-a (cadar a))
                          (pat-b (caar b))
                          (txt-b (cadar b)))
                     (cond
                      ((not (= groups-a groups-b))
                       (> groups-a groups-b))
                      ((not (= inputsize-a inputsize-b))
                       (<= inputsize-a inputsize-b))
                      ((not (string= pat-a pat-b))
                       (string< pat-a pat-b))
                      (else
                       (string< txt-a txt-b)))))))
    (sort traces sorter)))

;;;; Compare a set of matchers with string permutations
(define (compare-matchers-permutations matchers
                                       pat-abc pat-length
                                       txt-abc txt-length
                                       from amount)
  (let ((done 0)
        (to-do (min amount (* (perms (string-length pat-abc) 3 pat-length)
                              (perms (string-length txt-abc) 1 txt-length))))
        (old-procent-done 0)
        (lazy-patterns (string-permutations-lazy pat-abc 3 pat-length))
        (lazy-texts (string-permutations-lazy txt-abc 1 txt-length)))
    (letrec ((walk (lambda (pats txts output)
                     (cond
                      ((> done (+ from amount))
                       output)
                      ((lazy-null? pats)
                       output)
                      ((lazy-null? txts)
                       (walk (lazy-cdr pats) lazy-texts output))
                      (else
                       (set! done (1+ done))
                       (let ((procent-done (floor (* 100 (/ (- done from)
                                                            to-do)))))
                         (if (> procent-done old-procent-done)
                             (begin
                               (pn procent-done)
                               (set! old-procent-done procent-done))))
                       (let* ((pat (lazy-car pats))
                              ; append pattern to the end of text to
                              ; ensure there is always a match (again
                              ; to avoid special cases in morris-pratt and others)
                              (txt (string-append (lazy-car txts) pat)))
                         (walk pats (lazy-cdr txts)
                               (if (> done from)
                                   (if (= 0 (modulo done 20))
                                       (prune-traces
                                        (sort-traces
                                         (cons (compare-matchers matchers pat txt)
                                               output)))
                                       (cons (compare-matchers matchers pat txt)
                                             output))
                                   output))))))))
      (walk lazy-patterns lazy-texts '()))))

(define (compare-2-matchers matcher1 matcher2 lazy-patterns lazy-texts)
  (define (walk pats txts)
                     (cond
                      ((lazy-null? pats)
                       #t)
                      ((lazy-null? txts)
                       (walk (lazy-cdr pats) lazy-texts))
                      (else
                       (let* ((pat (lazy-car pats))
                              (txt (string-append (lazy-car txts) pat))
                              (r1 (matcher1 pat txt))
                              (r2 (matcher2 pat txt)))
                         (if (equal? r1 r2)
                             (walk pats (lazy-cdr txts))
                             (list (list pat txt) r1 r2))))))
  (walk lazy-patterns lazy-texts))

;;;; Compare a set of matchers with a set of pattern and texts
(define (compare-matchers-inputs matchers inputs)
  (define (walk inputs traces)
    (if (null? inputs)
        traces
        (let ((pat (caar inputs))
              (txt (cdar inputs)))
          (walk (cdr inputs)
                (prune-traces
                 (sort-traces
                  (cons (compare-matchers matchers pat txt) traces)))))))
  (walk inputs '()))

;;;; Prepare traces for output
(define t0 '(("A" "A") (((0) "1" "2" "3"))))
(define t1 '(("B" "B") (((0) "1" "2") ((1) "3"))))
(define t2 '(("C" "C") (((0) "1") ((1) "2" "3"))))
(define t3 '(("D" "D") (((0) "1") ((1) "2") ((2) "3"))))

(define t4 '(("e" "e") (((0) "1" "2") ((1) "3" "4"))))
(define t5 '(("f" "f") (((0) "1" "4") ((1) "2" "3"))))

(define ts0 (list t0))
(define ts1 (list t0 t1))
(define ts2 (list t0 t1 t2))
(define ts3 (list t0 t1 t2 t3))
(define ts4 (list t4 t5))

(define (pretty-traces traces)
  "types: (pattern = string, text = string, trace = int list)
 (((pattern * text) * (trace :: matcher list) list) list
   -> (matcher list * (pattern * text * trace) list) list

Sort traces into groups of matchers (that are trace equivalent)
along with a set of inputs and traces that uniquely identifies this
group apart from the other groups."

  (define (trace-matchers->tgroup trace trace-matchers)
    "tgroup format: ((matchers) ((pat txt (trace)) (pat2 txt2 (trace2))))"
    `(,(cdr trace-matchers) ((,@(car trace) ,(car trace-matchers)))))

  (define (init-tgroups trace)
    (list (apply append (map cdr (cadr trace)))
          '()))

  (define (trace-splits-group-amount trace group)
    (define (walk trace-matcherss group)
      (if (or (null? trace-matcherss) (null? group))
          0
          (let* ((cur-group (cdr (car trace-matcherss)))
                 (diff (lset-difference equal? group cur-group)))
            (+ (if (and (not (null? diff)) (not (equal? group diff))) 1 0)
               (walk (cdr trace-matcherss) diff)))))
    (walk (cadr trace) group))

  (define (trace-with-most-splits group traces)
    (define (walk traces max best)
      (if (null? traces)
          (and (> max 0) best)
          (let ((splits (trace-splits-group-amount (car traces) group)))
            (if (> splits max)
                (walk (cdr traces) splits (car traces))
                (walk (cdr traces) max best)))))
    (walk traces -1 '()))

  (define (split-tgroup-with-trace tgroup trace)
    (let ((obss (cadr tgroup)))
      (define (walk trace-matcherss group tgroups)
        (if (null? trace-matcherss)
            tgroups
            (let* ((trace-matchers (car trace-matcherss))
                   (split-group (lset-intersection equal? group
                                                   (cdr trace-matchers)))
                   (rest-group (lset-difference equal? group split-group))
                   (obs (caadr (trace-matchers->tgroup trace trace-matchers))))
              (if (null? split-group)
                  (walk (cdr trace-matcherss) group tgroups)
                  (walk (cdr trace-matcherss) rest-group
                        (cons (list split-group (cons obs obss))
                              tgroups))))))
      (walk (cadr trace) (car tgroup) '())))

  (define (split-tgroup tgroup traces)
    (let ((best-trace (trace-with-most-splits (car tgroup) traces)))
      (if (not best-trace)
          (list tgroup)
          (apply append
                 (map (lambda (tgroup) (split-tgroup tgroup traces))
                      (split-tgroup-with-trace tgroup best-trace))))))

  (map (lambda (tgroup)
         (if (null? (cadr tgroup))
             (cons (car tgroup)
                   (cdr (trace-matchers->tgroup (car traces)
                                                (caadar traces))))
             tgroup))
       (split-tgroup (init-tgroups (car traces)) traces)))

;;;; Output traces
(define (filter-specialized-from-tgroups tgroups)
  (filter (lambda (tgroup)
            (not (any (lambda (input-trace)
                        (null? (caddr input-trace)))
                      (cadr tgroup))))
          tgroups))

(define (sort-tgroups tgroups)
  (define (tgroups< a b)
    (if (= (length (cadr a)) (length (cadr b)))
        ((list<-maker inputs< equal?) (cadr a) (cadr b))
        (< (length (cadr a)) (length (cadr b)))))
  (define (inputs< a b)
    (define (input x) (string-append (car x) "   " (cadr x)))
    (define (trace x) (last x))
    (let ((str-a (input a))
          (str-b (input b))
          (trace-a (trace a))
          (trace-b (trace b)))
      (if (not (string= str-a str-b))
          (string< str-a str-b)
          (trace< trace-a trace-b))))
  (define trace< (list<-maker < =))
  (sort (map (lambda (tgroup)
               (list (sort (car tgroup) (list<-maker string< string=))
                     (reverse (cadr tgroup))))
             tgroups)
        tgroups<))

(define (output-traces-html-1 tgroups opts)
  (let ((get-pattern car)
        (get-text cadr)
        (get-trace caddr)
        (tgroups (sort-tgroups (filter-specialized-from-tgroups tgroups))))
    (define (get-opt name) (cdr (or (assoc name opts) '(0 . "N/A"))))
    (define (list->string xs)
      (cond
       ((null? xs) "")
       ((null? (cdr xs)) (format #f "~a" (car xs)))
       (else (format #f "~a, ~a" (car xs) (list->string (cdr xs))))))
    (define (list->string-br xs)
      (cond
       ((null? xs) "")
       ((null? (cdr xs)) (format #f "~a" (car xs)))
       (else (format #f "~a<br/> ~a" (car xs) (list->string-br (cdr xs))))))
    (define (p-tgroups tgroups)
      (format #f "
inputs: ~a<br>
pattern alphabet: ~a<br>
pattern length up to: ~a<br>
text alphabet: ~a<br>
text length up to: ~a<br><br>
The right column lists groups of matchers with the
same trace on all output.<br> The left column lists groups of input
and trace uniquely identifying the matchers in the right column.<br>
The left column is of the form: (pattern text) resulting trace.

<table border=2>~a</table>\n"
              (get-opt "inputs")
              (get-opt "patabc") (get-opt "patlength")
              (get-opt "txtabc") (get-opt "txtlength")
              (fold (lambda (g str) (string-append str (p-tgroup g)))
                    "" tgroups)))
    (define (p-tgroup tgroup)
      (format #f "<tr><td>\n<table border=1>~a</table>\n</td><td>~a</td></tr>\n"
              (fold (lambda (obs str)
                      (format #f "~a<tr><td>~a ~a</td><td>~a</td></tr>\n"
                              str
                              (get-pattern obs)
                              (get-text obs)
                              (list->string (get-trace obs))))
                    "" (cadr tgroup))
              (list->string-br (car tgroup))))
    (p-tgroups tgroups)))

(define (output-traces-tex tgroups opts)
  (let ((get-pattern car)
        (get-text cadr)
        (get-trace caddr)
        (tgroups (sort-tgroups (filter-specialized-from-tgroups tgroups))))
    (define (get-opt name) (cdr (or (assoc name opts) '(0 . "N/A"))))
    (define (list->string xs)
      (cond
       ((null? xs) "")
       ((null? (cdr xs)) (format #f "~a" (car xs)))
       (else (format #f "~a, ~a" (car xs) (list->string (cdr xs))))))
    (define (list->string-br xs)
      (cond
       ((null? xs) "")
       ((null? (cdr xs)) (format #f "~a" (car xs)))
       (else (format #f "~a<br/> ~a" (car xs) (list->string-br (cdr xs))))))
    (define (p-tgroups tgroups)
      (format #f "\\begin{tabular}{lll|l}\n
pattern & text & trace & matcher \\\\ \n
\\hline \\hline\n
~a\n\\end{tabular}\n"
              (fold (lambda (g str) (string-append str (p-tgroup g)))
                    "" tgroups)))
    (define (p-tgroup tgroup)
      (define (walk input-traces matchers)
        (if (and (null? input-traces) (null? matchers))
            "\\hline\n"
            (format #f "~a & ~a \\\\ \n~a"
                    (if (null? input-traces)
                        "&&"
                        (let ((input-trace (car input-traces)))
                          (format #f "~a & ~a & ~a" 
                                  (get-pattern input-trace)
                                  (get-text input-trace)
                                  (list->string (get-trace input-trace)))))
                    (if (null? matchers)
                        ""
                        (car matchers))
                    (walk (if (null? input-traces) '() (cdr input-traces) )
                          (if (null? matchers) '() (cdr matchers))))))
      (walk (cadr tgroup) (car tgroup)))
    (p-tgroups tgroups)))

;;;; Identify matcher

(define (identify-matcher-maker hook-begin hook-different-trace hook-end)
  (lambda (matcher matchers lazy-pats lazy-txts)
    (define (walk-matchers pat txt trace unknown eqv)
      (define (walk-matchers' unknown eqv new-unknown new-eqv)
          (cond
           ((pair? unknown)
            (let* ((matcher' (car unknown))
                   (trace' ((cadr matcher') pat txt)))
              (if (null? trace')
                  (walk-matchers' (cdr unknown)  eqv
                                  (cons matcher' new-unknown) new-eqv)
                  (walk-matchers' (cdr unknown) (cons matcher' eqv)
                                  new-unknown new-eqv))))
           ((pair? eqv)
            (let* ((matcher' (car eqv))
                   (trace' ((cadr matcher') pat txt)))
              (if (or (null? trace') (equal? trace trace'))
                  (walk-matchers' unknown (cdr eqv)
                                  new-unknown (cons matcher' new-eqv))
                  (begin (hook-different-trace pat txt trace matcher' trace')
                         (walk-matchers' unknown (cdr eqv)
                                         new-unknown new-eqv)))))
           (else
            (list new-unknown new-eqv))))
      (walk-matchers' unknown eqv '() '()))
    (define (walk pats txts unknown eqv)
      (cond
       ((or (lazy-null? pats) (and (null? unknown) (null? eqv)))
        (hook-end unknown eqv))
       ((lazy-null? txts)
        (walk (lazy-cdr pats) lazy-txts unknown eqv))
       (else
        (let* ((pat (lazy-car pats))
               (txt (string-append (lazy-car txts) pat))
               (trace ((cadr matcher) pat txt)))
          (if (null? trace)
              (walk pats (lazy-cdr txts) unknown eqv)
              (xapply (walk-matchers pat txt trace unknown eqv)
                      (lambda (unknown' eqv')
                        (walk pats (lazy-cdr txts) unknown' eqv'))))))))
    (hook-begin)
    (walk lazy-pats lazy-txts matchers '())))

(define (identify-matcher matcher matchers lazy-pats lazy-txts)
    (define (strings->string xs)
      (cond
       ((null? xs) "")
       ((null? (cdr xs)) (format #f "   ~a" (car xs)))
       (else (format #f "   ~a\n~a" (car xs) (strings->string (cdr xs))))))

  (define (hook-begin)
    (format #t "Comparing ~a against ~a other matchers...\n\n"
            (car matcher) (length matchers)))
  (define (hook-different-trace pat txt trace matcher' trace')
    (format #t (string-append "~a is different from ~a\n"
                              "   pattern '~a' and text '~a'\n"
                              "   ~a trace: ~a\n"
                              "   ~a trace: ~a\n\n")
            (car matcher) (car matcher') pat txt (car matcher) trace
            (car matcher') trace'))
  (define (hook-end unknown eqv)
    (format #t "No more inputs\n\n")
    (if (pair? unknown)
        (format #t "These matchers do not share any inputs with ~a:\n~a\n\n"
                (car matcher) (strings->string (map car unknown))))
    (if (pair? eqv)
        (format #t "The matcher ~a is trace equivalent with:\n~a\n\n"
                (car matcher) (strings->string (map car eqv)))
        (format #t "The matcher ~a is not trace equivalent with any matcher\n\n"
                (car matcher))))
  ((identify-matcher-maker hook-begin hook-different-trace hook-end)
   matcher matchers lazy-pats lazy-txts))

;; evolutionary tree generation
(define (matcher-ids matchers)
  (letrec ((walk (lambda (matchers i)
                   (if (null? matchers)
                       '()
                       (cons (cons (car matchers)
                                   (string-append "m" (number->string i)))
                             (walk (cdr matchers) (1+ i)))))))
    (walk (map car matchers) 0)))
                 
(define (dmatrix->phylip matchers dmatrix)
  (let ((matcher-ids (matcher-ids matchers)))
    (define (id matcher)
      (cdr (assoc matcher matcher-ids)))
    (define (walk dmatrix out)
      (if (null? dmatrix)
          out
          (walk (cdr dmatrix)
                (format #f (string-append "~a"
                                          "~10a~a\n")
                        out
                        (id (caar dmatrix))
                        (fold (lambda (i str)
                                (string-append str " "
                                               (number->string i)))
                              "" (map cdr (cdar dmatrix)))))))
    (format #f "~a\n~a"
            (length matcher-ids)
            (walk dmatrix ""))))

(define (pairwise-alignment gapcost diffcost s1 s2)
  (define (init-matrix n m)
    (let ((T (make-vector n)))
      (letrec ((walk (lambda (n)
                       (if (< n 0)
                           #t
                           (begin (vector-set! T n (make-vector m #f))
                                  (walk (- n 1)))))))
        (walk (- n 1))
        T)))
  (let ((matrix (init-matrix (+ 1 (string-length s1)) (+ 1 (string-length s2)))))
    (define (T i j)
      (vector-ref (vector-ref matrix i) j))
    (define (set-T i j value)
      (vector-set! (vector-ref matrix i) j value)
      value)
    (define (my-min . xs)
      (letrec ((walk (lambda (xs min-x)
                       (if (null? xs)
                           min-x
                           (walk (cdr xs) 
                                 (if (car xs)
                                     (if min-x
                                         (min (car xs) min-x)
                                         (car xs))
                                     min-x))))))
        (walk xs #f)))
    (define (d i j)
      (if (equal? (string-ref s1 (- i 1)) (string-ref s2 (- j 1)))
          0
          diffcost))
    (define (cost i j)
      (if (T i j)
          (T i j)
          (set-T i j
                 (my-min (and (> i 0) (> j 0)
                              (+ (cost (- i 1) (- j 1)) (d i j)))
                         (and (> i 0) (>= j 0)
                              (+ (cost (- i 1) j) gapcost))
                         (and (>= i 0) (> j 0)
                              (+ (cost i (- j 1)) gapcost))
                         (and (= i 0) (= j 0)
                              0)))))
    (cost (string-length s1) (string-length s2))))

(define (for-all-different-traces proc trace)
  (define (walk ms mstrace tms)
    (cond
     ((null? tms)
      #t)
     ((null? ms)
      (walk (cdar tms) (caar tms) (cdr tms)))
     (else
      (walk-m (car ms) mstrace (cdar tms) (caar tms) (cdr tms))
      (walk (cdr ms) mstrace tms))))
  (define (walk-m matcher mtrace ms mstrace tms)
    (cond
     ((and (null? tms) (null? ms))
      #t)
     ((null? ms)
      (walk-m matcher mtrace (cdar tms) (caar tms) (cdr tms)))
     (else
      (if (not (equal? mtrace mstrace))
          (proc matcher mtrace (car ms) mstrace))
      (walk-m matcher mtrace (cdr ms) mstrace tms))))
  (let ((tms (cadr trace)))
    (walk (cdar tms) (caar tms) (cdr tms))))

(define (trace-update-pairwise-alignment gapcost diffcost)
  (lambda (inc trace)
    (define (trace->string t)
      (list->string (map (lambda (i) (integer->char (+ 97 i))) t)))
    (let ((latest-key (cons "" ""))
          (latest-diff 0))
      (define (align st st')
        (let ((key (cons st st')))
          (if (equal? key latest-key)
              latest-diff
              (begin
                (let ((diff (pairwise-alignment gapcost diffcost st st')))
                  (set! latest-key key)
                  (set! latest-diff diff)
                  diff)))))
      (for-all-different-traces (lambda (m t m' t') 
                                  (let ((diff (align (trace->string t)
                                                     (trace->string t'))))
                                    (inc m m' diff)
                                    (inc m' m diff)))
                                trace))))

(define (trace-update-equal-or-not inc trace)
  (for-all-different-traces (lambda (m t m' t') (inc m m' 1) (inc m' m 1))
                            trace))

(define (distance-matrix trace-update matchers lazy-pats lazy-txts)
  (let ((matcher-names (map car matchers)))
    (define (init-dmatrix matchers)
      (map (lambda (m)
             (cons m (map (lambda (m)
                            (cons m 0))
                          matcher-names)))
           matcher-names))
    (let ((dmatrix (init-dmatrix matchers)))
      (define (inc-dmatrix from to with)
        (let ((entry (assoc to (cdr (assoc from dmatrix)))))
          (set-cdr! entry (+ with (cdr entry)))))
      (define (walk pats txts)
        (cond
         ((lazy-null? pats)
          #t)
         ((lazy-null? txts)
          (walk (lazy-cdr pats) lazy-txts))
         (else
          (let* ((pat (lazy-car pats))
                 (txt (string-append (lazy-car txts) pat)))
            (trace-update inc-dmatrix
                          (compare-matchers matchers pat txt))
            (walk pats (lazy-cdr txts))))))
      (walk lazy-pats lazy-txts)
      dmatrix)))

(define (rename-tree matchers tree)
  (let ((matcher-ids (matcher-ids matchers)))
    (letrec ((from-to-walk
              (lambda (matcher-ids out)
                (if (null? matcher-ids)
                    out
                    (from-to-walk (cdr matcher-ids)
                                  (cons (cons (cdar matcher-ids) 
                                              (caar matcher-ids))
                                        out))))))
      (string-replace-all-list tree (from-to-walk matcher-ids '())))))

(define (parse-tree tree)
  (let ((idx 0))
    (define (peek)
      (string-ref tree idx))
    (define (next)
      (set! idx (1+ idx))
      (string-ref tree (- idx 1)))
    (define (rewind)
      (set! idx (- idx 1)))
    (define (eof?)
      (>= idx (string-length tree)))
    (define (walk0)
      (next) (walk '()))
    (define (walk t)
      (case (next)
        ((#\newline #\space) (walk t))
        ((#\() (walk (cons (walk '()) t)))
        ((#\)) (walk-rparen t))
        ((#\,) (walk t))
        ((#\;) t)
        (else (rewind) (walk (cons (walk-name-value) t)))))
    (define (walk-rparen t)
      (case (next)
        ((#\newline #\space) (walk-rparen t))
        ((#\:) (cons (walk-value "") (reverse t)))
        (else (rewind) (reverse t))))
    (define (walk-name-value)
      (list (walk-name "") (walk-value "")))
    (define (walk-name acc)
      (let ((c (next)))
        (case c
          ((#\:) acc)
          (else (walk-name (string-append acc (char->string c)))))))
    (define (walk-value acc)
      (let ((c (next)))
        (case c
          ((#\) #\,) (rewind) (string->number acc))
          (else (walk-value (string-append acc (char->string c)))))))
    (walk0)))

(define (dot-parsed-tree ptree)
  (let ((hash-n 0)
        (hashs '()))
    (define (hash ptree)
      (let ((e (assoc ptree hashs)))
        (if e
            (cdr e)
            (new-hash ptree))))
    (define (new-hash ptree)
      (set! hashs (cons (cons ptree hash-n) hashs))
      (set! hash-n (1+ hash-n))
      (- hash-n 1))
    (define (node ptree)
      (format #f "n~a [label=\"~a\"]\n"
              (hash ptree)
              (if (string? (car ptree))
                  (car ptree)
                  "")))
    (define (edge from to)
      (format #f "n~a -> n~a\n"
              (hash from) (hash to)))
    (define (walk ptree dot)
      (if (string? (car ptree))
          (string-append (node ptree) dot)
          (fold (lambda (child dot)
                  (string-append (walk child dot)
                                 (edge ptree child)))
                (string-append dot (node ptree))
                (if (number? (car ptree))
                    (cdr ptree)
                    ptree))))
    (format #f "digraph {\n~a\n}\n"
            (walk ptree ""))))

(define (avg-trace-lengths matchers lazy-pats lazy-txts)
  (define (sort-car-strings ss)
    (sort ss (lambda (a b) (string< (car a) (car b)))))
  (define (init-trace-lengths matchers)
    (sort-car-strings (map (lambda (m) (cons (car m) 0)) matchers)))
  (define (update-trace trace-lengths trace)
    (map (lambda (m m-add) (cons (car m) (+ (cdr m) (cdr m-add))))
         trace-lengths
         (let ((text-length (string-length (cadar trace))))
           (sort-car-strings
            (apply append
                   (map (lambda (group)
                          (map (lambda (m) (cons m (/ (length (car group))
                                                      text-length)))
                               (cdr group)))
                        (cadr trace)))))))
  (define (divide-trace-lengths trace-lengths divider)
    (map (lambda (mt) (cons (car mt) (/ (cdr mt) divider)))
         trace-lengths))
  (let ((inputs 0))
    (define (1+inputs) (set! inputs (1+ inputs)))
    (define (walk trace-lengths pats txts)
      (cond
       ((lazy-null? pats)
        trace-lengths)
       ((lazy-null? txts)
        (walk trace-lengths (lazy-cdr pats) lazy-txts))
       (else
        (let* ((pat (lazy-car pats))
               (txt (string-append (lazy-car txts) pat)))
          (1+inputs)
          (walk (update-trace trace-lengths
                              (compare-matchers matchers pat txt))
                pats (lazy-cdr txts))))))
    (sort (divide-trace-lengths 
           (walk (init-trace-lengths matchers) lazy-pats lazy-txts)
           inputs)
          (lambda (a b) (< (cdr a) (cdr b))))))
  

"comparator.scm"
