;; framework_composition.scm
;; 
;; This file defines functions used to compose string matchers.

;;;; cache
(define pos? char?)

(define (neg? e)
  (or (null? e) (pair? e)))

(define (cache-ref cache j)
  "Returns the information in the given cache at the text indice j. If
no information is found the empty list (empty set of negative
informmation) is returned."
  (letrec ((walk (lambda (cache fcache neg)
                   (cond
                    ((and (null? cache) (null? fcache))
                     neg)
                    ((null? fcache)
                     (walk (cdr cache) (car cache) neg))
                    (else
                     (let ((j' (caar fcache))
                           (info (cdar fcache)))
                       (cond
                        ((not (= j j'))
                         (walk cache (cdr fcache) neg))
                        ((pos? info)
                         info)
                        ((neg? info)
                         (walk cache (cdr fcache) (append neg info))))))))))
    (walk cache '() '())))

(define (cache-hit? cache j pat-char)
  "Whether the given cache has information about a text indice j and a
character pat-char."
  (let ((info (cache-ref cache j)))
    (or (pos? info)
        (and (neg? info) (member pat-char info)))))

(define (cache-mismatch? cache j pat-char)
  "Whether the cache can tell us if the text indice j is different
from the character pat-char."
  (let ((info (cache-ref cache j)))
    (if (pos? info)
        (not (equal? info pat-char))
        (member pat-char info))))

(define (cache-upd-pos cache j pat-char)
  (cons (cons (cons j pat-char) (car cache))
        (cdr cache)))

(define (cache-upd-neg cache j pat-char)
  (let* ((info (cdr (or (assoc j (car cache)) '(())))))
    (if (pos? info)
        cache
        (cons (cons (cons j (cons pat-char info)) (car cache))
              (cdr cache)))))

(define (cache-filter f cache)
  (letrec ((walk-flat (lambda (fcache)
                        (if (null? fcache)
                            '()
                            (if (f (cdar fcache))
                                (cons (car fcache) 
                                      (walk-flat (cdr fcache)))
                                (walk-flat (cdr fcache))))))
           (walk (lambda (cache)
                   (if (null? cache)
                       '()
                       (cons (walk-flat (car cache))
                             (walk (cdr cache)))))))
    (walk cache)))

(define (cache-shifts pattern txt-offset cache)
  "pattern * txt-offset * cache -> shifts

Calculate the number of shifts possible with a given cache on a
given pattern and offset."
  (letrec ((walk (lambda (i shifts)
                   (cond
                    ((>= i (string-length pattern))
                     shifts)
                    ((cache-mismatch? cache 
                                      (+ txt-offset shifts i)
                                      (string-ref pattern i))
                     (walk 0 (1+ shifts)))
                    (else
                     (walk (1+ i) shifts))))))
    (walk 0 1)))

;;; cache pruners
(define (prune-none cache)
  cache)

(define (prune-all cache)
  '())

(define (prune-pos cache)
  (cache-filter (lambda (e) (not (pos? e))) cache))

(define (prune-neg cache)
  (cache-filter (lambda (e) (not (neg? e))) cache))

(define (prune-pos-older-than i)
  (lambda (cache)
    (append (list-head' cache i)
            (prune-pos (list-tail' cache i)))))

(define (prune-neg-older-than i)
  (lambda (cache)
    (append (list-head' cache i)
            (prune-neg (list-tail' cache i)))))

(define (prune-older-than i)
  (lambda (cache)
    (list-head' cache i)))

;;;; order
(define (order-left-to-right pat-length)
  (range-up 0 pat-length))

(define (order-right-to-left pat-length)
  (range-down 0 pat-length))

(define (order-last-left-to-right pat-length)
  (cons (- pat-length 1)
        (range-up 0 (- pat-length 1))))

(define (order-last-only pat-length)
  (list (- pat-length 1)))

(define (order-last-two pat-length)
  (if (= pat-length 1)
      (list 0)
      (list (- pat-length 2) (- pat-length 1))))

(define (order-second-only pat-length)
  (list (if (= pat-length 1) 0 1)))

(define (order-left-to-right-skip-second pat-length)
  (cons 0 (range-up 2 pat-length)))

(define (order-third-to-right-first pat-length)
  (range-up-aux 2 pat-length (list 0)))

(define (order-secondlast-last-right-to-left pat-length)
  (if (= 1 pat-length)
      (list 0)
      (cons (- pat-length 2) 
            (cons (- pat-length 1)
                  (range-down 0 (max 0 (- pat-length 2)))))))

(define (order-last-first-middle-rest pat-length)
  (case pat-length
    ((1) (list 0))
    ((2) (list 1 0))
    (else
     (let ((first 0)
           (middle (floor (/ pat-length 2)))
           (last (- pat-length 1)))
       (cons last 
             (cons first 
                   (cons middle 
                         (range-up-aux (1+ first) middle 
                                       (range-up (1+ middle) last)))))))))

;;;; match makers

(define (match-general orderer pruner is-table stop-on-mismatch)
  (lambda (pattern text txt-offset cache)
    (letrec ((walk (lambda (pat-indices cache trace)
                     (cond
                      ((null? pat-indices)
                       (list #t cache trace -1))
                      ((or (>= (car pat-indices) (string-length pattern))
                           (>= (+ (car pat-indices) txt-offset)
                               (string-length text)))
                       (list #f cache trace -1))
                      (else
                       (let* ((i (car pat-indices))
                              (pat-char (string-ref pattern i))
                              (j (+ txt-offset i))
                              (txt-char (string-ref text j))
                              (trace' (if (cache-hit? cache j pat-char)
                                          trace (cons j trace))))
                         (if (equal? pat-char txt-char)
                             (walk (cdr pat-indices)
                                   (cache-upd-pos cache j pat-char)
                                   trace')
                               (let ((cache' 
                                      (if is-table 
                                          (cache-upd-pos cache j txt-char)
                                          (cache-upd-neg cache j pat-char))))
                                 (if stop-on-mismatch
                                     (list #f cache' trace' i)
                                     (walk (cdr pat-indices) cache' 
                                           trace'))))))))))
      (xapply (walk (orderer (string-length pattern)) (cons '() cache) '())
              (lambda (match? cache trace mismatch-i)
                (let ((cache' (pruner cache)))
                  (list match?
                        (cache-shifts pattern txt-offset cache')
                        cache' (reverse trace) mismatch-i)))))))

(define (match-basic-or-table orderer pruner is-table)
  "match-basic-or-table orderer pruner is-table
orderer * pruner * boolean ->
  pattern * text * txt-offset * cache ->
    match? * shifts * cache * trace * mismatch-i"
  (match-general orderer pruner is-table #t))

(define (match-basic-full orderer pruner)
  (match-general orderer pruner #f #f))

(define (match-table-full orderer pruner)
  (match-general orderer pruner #t #f))

(define (match-basic orderer pruner)
  "Returns a match function. A match function checks if a pattern is
the prefix of a given text suffix and returns:
 - the result
 - how much it is safe to shift next text suffix
 - the changed cache
 - a trace and which pattern indice the matching failed on"
  (match-basic-or-table orderer pruner #f))

(define (match-table orderer pruner)
  "Like match-basic but emulates using a bad character shift
heuristics table."
  (match-basic-or-table orderer pruner #t))

(define (match-shifts match pruner)
  (lambda (pattern text txt-offset cache)
    (xapply (match pattern text txt-offset '())
            (lambda (match? shifts cache' trace mismatch-i)
              (let ((cache'' (pruner (append cache' cache))))
                (list match? 
                      (cache-shifts pattern txt-offset cache'')
                      cache'' trace mismatch-i))))))

(define (match-basic-shifts orderer pruner)
  "Like match-basic but doesn't use the cache to potentially skip over
some indices during matching."
  (match-shifts (match-basic orderer pruner) pruner))

(define (match-table-shifts orderer pruner)
  "Like match-table but doesn't use the cache to potentially skip over
some indices during matching."
  (match-shifts (match-table orderer pruner) pruner))

(define (match-table-full-shifts orderer pruner)
  (match-shifts (match-table-full orderer pruner) pruner))

(define (match-backtracking match1 match2)
  "Combines 2 match functions: match1 and match2. If match1 fails,
shifts the maximum of match1 and match2. Only match1 is traced (needs
to be a working string matcher)."
  (lambda (pattern text txt-offset caches)
    (let ((cache1 (if (pair? caches) (car caches) '()))
          (cache2 (if (pair? caches) (cadr caches) '())))
      (xapply (match1 pattern text txt-offset cache1)
              (lambda (match?1 shifts1 cache1' trace1 mismatch-i1)
                (if match?1
                    (list match?1
                          shifts1 
                          (list cache1' cache2) 
                          trace1 
                          mismatch-i1)
                    (xapply (match2 pattern text txt-offset cache2)
                            (lambda (match?2 shifts2 cache2' trace2 mismatch-i2)
                              (list match?1
                                    (max shifts1 shifts2)
                                    (list cache1' cache2') 
                                    (append trace1 trace2)
                                    mismatch-i1)))))))))

(define (match-sequential match1 match2)
  "Combines 2 match functions: match1 match2. If match1 succeeds,
continue with match2 and shifts the maximum of match1 and
match2. match1 and match2 together needs to be a working string
matcher."
  (lambda (pattern text txt-offset caches)
    (let ((cache1 (if (pair? caches) (car caches) '()))
          (cache2 (if (pair? caches) (cadr caches) '())))
      (xapply (match1 pattern text txt-offset cache1)
              (lambda (match?1 shifts1 cache1 trace1 mismatch-i1)
                (if match?1
                    (xapply (match2 pattern text txt-offset cache2)
                            (lambda (match?2 shifts2 cache2 trace2 mismatch-i2)
                              (list match?2
                                    (max shifts1 shifts2)
                                    (list cache1 cache2)
                                    (append trace1 trace2)
                                    mismatch-i2)))
                    (list match?1 
                          shifts1 
                          (list cache1 cache2) 
                          trace1 
                          mismatch-i1)))))))

(define (match-parallel match1 match2)
  "Combines 2 match functions: match1 match2. Returns the best result
of match1 or match2. Either match1 or match2 is traced (needs to be a
working string matcher)."
  (lambda (pattern text txt-offset caches)
    (let ((cache1 (if (pair? caches) (car caches) '()))
          (cache2 (if (pair? caches) (cadr caches) '())))
      (xapply (match1 pattern text txt-offset cache1)
              (lambda (match?1 shifts1 cache1 trace1 mismatch-i1)
                (xapply (match2 pattern text txt-offset cache2)
                        (lambda (match?2 shifts2 cache2 trace2 mismatch-i2)
                          (list (and match?1 match?2)
                                (max shifts1 shifts2)
                                (list cache1 cache2)
                                (append trace1 trace2)
                                (if (>= shifts1 shifts2) mismatch-i1 mismatch-i2)
                                ))))))))

(define (match-alternate match1 match2)
  "Combines 2 match functions: match1 match2. If match1 fails,
run match2 starting from where match1 shifts to. Only match1
is traced (needs to be a working string matcher)."
  (lambda (pattern text txt-offset caches)
    (let ((cache1 (if (pair? caches) (car caches) '()))
          (cache2 (if (pair? caches) (cadr caches) '())))
      (xapply (match1 pattern text txt-offset cache1)
              (lambda (match?1 shifts1 cache1 trace1 mismatch-i1)
                (if match?1
                    (list match?1 
                          shifts1 
                          (list cache1 cache2) 
                          trace1 
                          mismatch-i1)
                    (xapply (match2 pattern text (+ txt-offset shifts1) cache2)
                            (lambda (match?2 shifts2 cache2 trace2 mismatch-i2)
                              (list #f
                                    (if match?2 
                                        shifts1 
                                        (+ shifts1 shifts2))
                                    (list cache1 cache2) 
                                    (append trace1 trace2) 
                                    mismatch-i1)))))))))

(define (match-skew match1 match2)
  "Combines 2 match functions: match1 match2. If match1 fails,
run match2 on the pattern ending just after where match1 failed. Only
match1 is traced (needs to be a working string matcher)."
  (lambda (pattern text txt-offset caches)
    (let ((cache1 (if (pair? caches) (car caches) '()))
          (cache2 (if (pair? caches) (cadr caches) '()))
          (pl (string-length pattern)))
      (xapply (match1 pattern text txt-offset cache1)
              (lambda (match?1 shifts1 cache1 trace1 mismatch-i1)
                (if match?1
                    (list match?1
                          shifts1
                          (list cache1 cache2) 
                          trace1
                          mismatch-i1)
                    (let ((txt-offset2 (- (+ 1 txt-offset mismatch-i1) pl)))
                      (xapply (match2 pattern text txt-offset2 cache2)
                              (lambda (match?2 shifts2 cache2 trace2 mismatch-i2)
                                (list #f
                                      (max shifts1 (- shifts2 (- txt-offset 
                                                                 txt-offset2)))
                                      (list cache1 cache2) 
                                      (append trace1 trace2)
                                      mismatch-i2))))))))))

(define (match-fail)
  "Always fails. Or, a match function that always shifts by 1."
  (lambda (pattern text txt-offset cache)
    (list #f 1 '() '() -1)))

(define (dont-trace match)
  "Wrapper for match functions that disable tracing."
  (lambda (pattern text txt-offset cache)
    (xapply (match pattern text txt-offset cache)
            (lambda (match? shifts cache' trace mismatch-i)
              (list match? shifts cache' '() mismatch-i)))))

(define (dont-trace-last-index-unless-matching match)
  "Wrapper for match functions that remove tracing of the last pattern
index (used in the smith matcher)."
  (lambda (pattern text txt-offset cache)
    (xapply (match pattern text txt-offset cache)
            (lambda (match? shifts cache' trace mismatch-i)
              (if match?
                  (list match? shifts cache' trace mismatch-i)
                  (list match? shifts cache' 
                        (filter (lambda (j) 
                                  (not (= j (+ txt-offset 
                                               (- (string-length pattern) 1)))))
                                trace) mismatch-i))))))

(define (trace-remove-duplicates trace)
  (letrec ((walk (lambda (trace seen)
                   (if (null? trace)
                       '()
                       (let ((idx (car trace)))
                         (if (member idx seen)
                             (walk (cdr trace) seen)
                             (cons idx (walk (cdr trace) (cons idx seen)))))))))
    (walk trace '())))
                       

(define (make-matcher match)
  "match -> (pattern * text -> trace)

Turns a match function into a string matcher taking a pattern and a
text and returning a trace."
  (lambda (pattern text)
    (letrec ((rec (lambda (txt-offset cache trace)
                    (xapply (match pattern text txt-offset cache)
                            (lambda (match? shifts cache trace' mismatch-i)
                              (let ((txt-offset' (+ txt-offset shifts))
                                    (trace'' (append trace 
                                                     (trace-remove-duplicates 
                                                      trace'))))
                                (if (or match?
                                        (> (+ txt-offset' 
                                              (string-length pattern))
                                           (string-length text)))
                                    trace''
                                    (rec txt-offset' cache trace''))))))))
      (cond
       ((= 0 (string-length pattern))
        '())
       ((> (string-length pattern) (string-length text))
        '())
       (else
        (rec 0 '() '()))))))

"framework.scm"
