;; identify_matcher.scm
;;
;; This file is an CLI executable command.  
;;
;; It compares a single matcher against a set of other matchers over a
;; given set of inputs.

(use-modules (ice-9 pretty-print))
(use-modules (srfi srfi-37))

(load (string-append (getcwd) "/framework/init.scm"))

(define (p-stderr . args)
  (if (null? args)
      #t
      (let ((stderr (current-error-port)))
        (write (car args) stderr) 
        (display " " stderr) 
        (apply p (cdr args)))))

(define (pn-stderr . args)
  (apply p-stderr args)
  (newline (current-error-port)))

(define pat-abc "ab")
(define pat-length 4)
(define txt-abc "abc")
(define txt-length 5)

(define (load-matcher file)
  (let ((matchers (load (if (eq? #\/ (string-ref file 0))
                           file
                           (string-append (dirname file) "/../" 
                                          (basename file))))))
    (verify-matchers matchers)
    matchers))

(define (verify-matchers matchers)
  (let ((err (lambda (m error)
               (error "Error in matchers entry '" m "': " error))))
    (for-each (lambda (m) 
                (cond
                 ((not (and (= (length m) 2) 
                            (string? (car m))
                            (procedure? (cadr m))))
                  (err m (string-append "Needs to be of the form "
                                        "(\"matcher-name\" procedure).")))
                 ((not (list? ((cadr m) "a" "a")))
                  (err m (string-append "Procedure returned non-list output, "
                                        "needs to return a lst of integers.")))
                 (else
                  #t)))
              matchers)))

(define (get-opts-traces assoc-or actual-args)
  (let* ((matchers (load-matcher (car actual-args)))
         (matcher-name (cadr actual-args))
         (matcher (find (lambda (m) (equal? (car m) matcher-name)) 
                        matchers))
         (matchers' (filter (lambda (m) (not (equal? (car m) matcher-name)))
                            matchers)))
    (if (not matcher)
        (error "Matcher not found in MATCHERS-FILE.")
        (let* ((pat (assoc-or "pat" #f))
               (patabc (assoc-or "patabc" pat-abc))
               (patlength (assoc-or "patlength" pat-length))
               (txtabc (assoc-or "txtabc" txt-abc))
               (txtlength (assoc-or "txtlength" txt-length))
               (lpats (if pat
                          (list->lazy (list pat))
                          (string-permutations-lazy patabc 3 patlength)))
               (ltxts (string-permutations-lazy txtabc 1 txtlength)))
          (identify-matcher matcher matchers' lpats ltxts)))))

(define (main args)
  (let* ((name-cons (lambda (name)
                      (lambda (opt name' arg loads) 
                        (cons (cons name arg) loads))))
         (name-cons-int (lambda (name)
                      (lambda (opt name' arg loads) 
                        (cons (cons name (string->number arg)) loads))))
         (opts (args-fold 
                (cdr args)
                (list 
                 (option '(#\h "help") #f #f (name-cons "help"))
                 (option '(#\p "pat") #t #f (name-cons "pat"))
                 (option '(#\a "patabc") #t #f (name-cons "patabc"))
                 (option '(#\l "patlength") #t #f (name-cons-int "patlength"))
                 (option '(#\A "txtabc") #t #f (name-cons "txtabc"))
                 (option '(#\L "txtlength") #t #f (name-cons-int "txtlength"))
                 )
                (lambda (opt name arg loads)
                  (error (format #f "Unrecognized option `~A'" name)))
                (lambda (arg loads) (cons (cons "arg" arg) loads))
                '()))
         (actual-args (reverse (map cdr (filter (lambda (opt) 
                                                  (equal? (car opt) "arg"))
                                                opts))))
         (number-of-args (length actual-args)))
    (cond
     ((or (assoc "help" opts) (not (= 2 number-of-args)))
      (display-help))
     (else
      (let* ((assoc-or (lambda (name or-val)
                         (let ((opt (assoc name opts)))
                           (cond
                            ((not opt) or-val)
                            ((not (cdr opt)) #t)
                            (else (cdr opt)))))))
        (get-opts-traces assoc-or actual-args))))))

(define (display-help)
  (display "Usage: program [OPTIONS] MATCHERS-FILE MATCHER
Compares MATCHER against string matchers from MATCHERS-FILE.

  -h, --help                print this message

  -p, --pat=STRING          use only this pattern
  -a, --patabc=ALPHABET     the pattern alphabet, default is ab
  -l, --patlength=SIZE      the pattern lengths, from 3 to SIZE, default is 3
  -A, --txtabc=ALPHABET     the text alphabet, default is abc
  -L, --txtlength=SIZE      the text length, default is 4


EXAMPLE: Identify Knuth-Morris-Pratt
  Most basic usage
    ./identify_matcher matchers.scm fw_kmp
"))

    
(main (command-line))
