(use-modules (ice-9 popen))
(use-modules (ice-9 rdelim))
(use-modules (system foreign))
(use-modules (rnrs bytevectors))
(use-modules (srfi srfi-1))
(use-modules (srfi srfi-37))

;;;; list paths
(set! %load-path (cons "." %load-path))

(define (init)
  "Load scheme files containing the framework."
  (load "framework_composition.scm")
  (load "framework_comparison.scm")
  (load "matcher_definitions.scm")
  (load "matchers.scm"))

;;;; helper functions
(define (p . args)
  (if (null? args)
      #t
      (begin (write (car args)) (display " ") (apply p (cdr args)))))

(define (pn . args)
  (apply p args)
  (newline)
  (last args))

(define (range-up-aux from to lst)
  (letrec ((walk (lambda (i)
                   (if (>= i to)
                       lst
                       (cons i (walk (1+ i)))))))
    (walk from)))
(define (range-up from to)
  "Returns the list of integers [from, from+1, ..., to-1]."
  (range-up-aux from to '()))

(define (range-down from to)
  "Returns the list of integers [to-1, to-2, ..., from]."
  (letrec ((walk (lambda (i)
                   (if (< i from)
                       '()
                       (cons i (walk (- i 1)))))))
    (walk (- to 1))))

(define (list-head' list k)
  (if (or (null? list) (<= k 0))
      '()
      (cons (car list) (list-head' (cdr list) (- k 1)))))

(define (list-tail' list k)
  (cond
   ((null? list)
    '())
   ((<= k 0)
    list)
   (else
    (list-tail' (cdr list) (- k 1)))))

(define (xapply arg fun)
  "(apply) with reversed argument order."
  (apply fun arg))

(define (perms asize ssize-from ssize-to)
  "Number of permutations for strings with alphabet size ASIZE and
lenghts from SSIZE-FROM up to SSIZE-TO."
  (if (< ssize-to ssize-from)
      0
      (+ (expt asize ssize-to) (perms asize ssize-from (- ssize-to 1)))))

(define (exec command)
  (let ((port (open-input-pipe command)))
    (letrec ((walk (lambda (out)
                     (let ((line (read-line port)))
                       (if (eof-object? line)
                           out
                           (walk (string-append out line "\n")))))))
      (walk ""))))

(define (my-time f)
  (let ((start (times)))
    (f)
    (let ((end (times)))
      (display (format (string-append "~7@a ms cpu time\n"
                                      "~7@a ms real time\n")
                       (* 10 (- (tms:utime end) (tms:utime start)))
                       (* 10 (- (tms:clock end) (tms:clock start))))))))

(define (char->string char)
  (make-string 1 char))

(define (string-trim string trims)
  (cond
   ((= 0 (string-length string))
    string)
   ((string? trims)
    (string-trim string (string->list trims)))
   ((member (string-ref string 0) trims)
    (string-trim (substring string 1 (string-length string)) trims))
   ((member (string-ref string (- (string-length string) 1)) trims)
    (string-trim (substring string 0 (- (string-length string) 1)) trims))
   (else
    string)))

(define (string-tokenize string splitters)
  (letrec ((loop (lambda (i token)
                   (if (>= i (string-length string))
                       (if (string=? token "")
                           '()
                           (cons token '()))
                       (let ((c (string-ref string i)))
                         (if (member c splitters)
                             (cons token (loop (1+ i) ""))
                             (loop (1+ i) 
                                   (string-append token (char->string c)))))))))
    (if (string? splitters)
        (string-tokenize string (string->list splitters))
        (loop 0 ""))))

(define (my-write-file filename obj)
  (let ((port (open-file filename "w")))
    (write obj port)
    (close-port port)
    obj))

(define (my-display-file filename obj)
  (let ((port (open-file filename "w")))
    (display obj port)
    (close-port port)
    obj))

(define (my-read-file filename)
  (let* ((port (open-file filename "r"))
         (output (read port)))
    (close-port port)
    output))

(define my-read-stdin read)

(define (my-read-file-as-string filename)
  (let ((port (open-file filename "r")))
    (define (walk output)
      (let ((c (read-char port)))
        (if (eof-object? c)
            (list->string (reverse output))
            (walk (cons c output)))))
    (let ((output (walk '())))
      (close-port port)
      output)))

(define (my-command-line)
  (command-line))

;;;; External matchers (executable programs)
(define (run-external command args)
  (exec (string-append command " " args)))

(define (trace-string-match command pattern text)
  (let* ((output (run-external command (string-append pattern " " text))))
    (map string->number
         (string-tokenize (string-trim output "\n ") " "))))

(define (make-external-matcher command)
  (lambda (pattern text)
    (trace-string-match command pattern text)))

;;;; Dynamic matchers (shared libraries)
(define (make-dynamic-matcher so-file matcher-name)
  (let* ((dyn-obj (dynamic-link so-file))
         (matcher (pointer->procedure
                   '*
                   (dynamic-func matcher-name dyn-obj)
                   (list '* int '* int))))
    (letrec ((get-next (lambda (ptr) (parse-c-struct ptr (list int '*))))
             (get-trace 
              (lambda (ptr)
                (if (equal? ptr %null-pointer)
                    '()
                    (xapply (get-next ptr)
                            (lambda (element next)
                              (cons element (get-trace next)))))))
             (string->c (lambda (s) 
                          (let* ((len (string-length s))
                                 (bv-nils (make-bytevector (+ 4 len) 0))
                                 (bv-s (string->utf8 s)))
                            ;; some matchers seem to need more than one
                            ;; null terminator (ie. berry_ravindran).
                            (bytevector-copy! bv-s 0 bv-nils 0 len)
                            (bytevector->pointer bv-nils)))))
      (lambda (pattern text)
        (let ((pattern-ptr (string->c pattern))
              (text-ptr (string->c text)))
          (get-trace (matcher pattern-ptr (string-length pattern) 
                              text-ptr (string-length text))))))))

(define (string-replace-all string from to)
  (let ((froml (string-length from))
        (tol (string-length to)))
    (letrec ((walk (lambda (string index)
                     (let ((match-index (string-contains string from index)))
                       (if match-index
                           (walk (string-replace string to match-index
                                                 (+ match-index froml))
                                 (+ match-index tol))
                           string)))))
      (walk string 0))))
(define (string-replace-all-list string from-tos)
  (fold (lambda (from-to string)
          (string-replace-all string (car from-to) (cdr from-to)))
        string from-tos))


(init)

"init.scm"
