;; table_separating_matchers.scm
;;
;; This file is an CLI executable command.
;; It generates tables separating string matchers.

(use-modules (ice-9 pretty-print))
(use-modules (srfi srfi-37))

(load (string-append (getcwd) "/framework/init.scm"))

(define (p-stderr . args)
  (if (null? args)
      #t
      (let ((stderr (current-error-port)))
        (write (car args) stderr) 
        (display " " stderr) 
        (apply p (cdr args)))))

(define (pn-stderr . args)
  (apply p-stderr args)
  (newline (current-error-port)))

(define pat-abc "ab")
(define pat-length 3)
(define txt-abc "abc")
(define txt-length 4)
(define default-output "html")

(define (load-matcher file)
  (let ((matchers (load (if (eq? #\/ (string-ref file 0))
                           file
                           (string-append (dirname file) "/../" 
                                          (basename file))))))
    (verify-matchers matchers)
    matchers))

(define (read-input-file file)
  (let* ((trims "\r\n\t ")
         (not-empty-string? (lambda (x) (not (equal? x ""))))
         (contents (string-trim (my-read-file-as-string file) trims)))
    (map (lambda (x) (let ((input (filter not-empty-string? 
                                          (string-tokenize 
                                           (string-trim x trims) " \t"))))
                       (cons (car input) (cadr input))))
                 (string-tokenize contents "\n"))))

(define (verify-matchers matchers)
  (let ((err (lambda (m error)
               (error "Error in matchers entry '" m "': " error))))
    (for-each (lambda (m) 
                (cond
                 ((not (and (= (length m) 2) 
                            (string? (car m))
                            (procedure? (cadr m))))
                  (err m (string-append "Needs to be of the form "
                                        "(\"matcher-name\" procedure).")))
                 ((not (list? ((cadr m) "a" "a")))
                  (err m (string-append "Procedure returned non-list output, "
                                        "needs to return a lst of integers.")))
                 (else
                  #t)))
              matchers)))

(define (get-opts-traces assoc-or actual-args)
  (let ((raw-input (assoc-or "raw-input" #f))
        (input-file (assoc-or "input-file" #f)))
    (if raw-input
        (let ((first-file-opts-traces (my-read-file (car actual-args))))
          (list (car first-file-opts-traces)
                (fold (lambda (file traces)
                        (append traces (cadr (my-read-file file))))
                      (cadr first-file-opts-traces) (cdr actual-args))))
        (let* ((matchers (load-matcher (car actual-args))))
          (if input-file
              (let ((inputs (read-input-file input-file)))
                (list `(("inputs" . ,inputs))
                      (compare-matchers-inputs matchers inputs)))
              (let* ((patabc (assoc-or "patabc" pat-abc))
                     (patlength (assoc-or "patlength" pat-length))
                     (txtabc (assoc-or "txtabc" txt-abc))
                     (txtlength (assoc-or "txtlength" txt-length))
                     (input-size (* (perms (string-length patabc) 3 patlength)
                                    (perms (string-length txtabc) 1 txtlength)))
                     (chunks (assoc-or "chunks" #f))
                     (chunk (and chunks (assoc-or "chunk" #f)))
                     (chunk (and chunk (- chunk 1)))
                     (from (if (and chunk chunks)
                               (* chunk (floor (/ input-size chunks)))
                               0))
                     (amount (if (and chunk chunks)
                                 (floor (/ input-size chunks))
                                 input-size)))
                (set! pn (if (assoc-or "no-progress" #f) 
                             (lambda args #t) pn-stderr))
                (list `(("patabc" . ,patabc) 
                        ("patlength" . ,patlength)
                        ("txtabc" . ,txtabc) 
                        ("txtlength" . ,txtlength))
                      (compare-matchers-permutations matchers 
                                                     patabc patlength 
                                                     txtabc txtlength
                                                     from amount))))))))

(define (main args)
  (let* ((name-cons (lambda (name)
                      (lambda (opt name' arg loads) 
                        (cons (cons name arg) loads))))
         (name-cons-int (lambda (name)
                      (lambda (opt name' arg loads) 
                        (cons (cons name (string->number arg)) loads))))
         (opts (args-fold 
                (cdr args)
                (list 
                 (option '(#\h "help") #f #f (name-cons "help"))
                 (option '(#\c "chunk") #t #f (name-cons-int "chunk"))
                 (option '(#\C "chunks") #t #f (name-cons-int "chunks"))
                 (option '(#\a "patabc") #t #f (name-cons "patabc"))
                 (option '(#\l "patlength") #t #f (name-cons-int "patlength"))
                 (option '(#\A "txtabc") #t #f (name-cons "txtabc"))
                 (option '(#\L "txtlength") #t #f (name-cons-int "txtlength"))
                 (option '(#\p "no-progress") #f #f (name-cons "no-progress"))
                 (option '(#\o "output") #t #f (name-cons "output"))
                 (option '(#\i "input-file") #t #f (name-cons "input-file"))
                 (option '(#\r "raw-input") #f #f (name-cons "raw-input"))
                 )
                (lambda (opt name arg loads)
                  (error (format #f "Unrecognized option `~A'" name)))
                (lambda (arg loads) (cons (cons "arg" arg) loads))
                '()))
         (actual-args (reverse (map cdr (filter (lambda (opt) 
                                                  (equal? (car opt) "arg"))
                                                opts))))
         (number-of-args (length actual-args)))
    (cond
     ((or (assoc "help" opts) (null? actual-args))
      (display-help))
     (else
      (let* ((assoc-or (lambda (name or-val)
                         (let ((opt (assoc name opts)))
                           (cond
                            ((not opt) or-val)
                            ((not (cdr opt)) #t)
                            (else (cdr opt))))))
             (opts-traces (get-opts-traces assoc-or actual-args))
             (opts (car opts-traces))
             (traces (cadr opts-traces))
             (output (assoc-or "output" default-output)))
          (cond 
           ((equal? output "raw") (pretty-print (list opts traces)))
           ((equal? output "html") (display (output-traces-html-1 
                                             (pretty-traces 
                                              (sort-traces traces)) opts)))
           ((equal? output "tex") (display (output-traces-tex 
                                            (pretty-traces 
                                             (sort-traces traces)) opts)))
           (else (error "Error! invalid --output type"))))))))

(define (display-help)
  (display "Usage: program [OPTIONS] FILE
Compares string matchers from FILE on string permutations.

  -h, --help                print this message

  -a, --patabc=ALPHABET     the pattern alphabet, default is ab
  -l, --patlength=SIZE      the pattern lengths, from 3 to SIZE, default is 3
  -A, --txtabc=ALPHABET     the text alphabet, default is abc
  -L, --txtlength=SIZE      the text length, default is 4

  -i, --input-file=FILE     get pattern/text inputs from file

  -r, --raw-input           reads arguments as files having --output=raw data

  -c, --chunk=CHUNK         which chunk of inputs to process, values [1,chunks]
  -C, --chunks=SIZE         how many chunks to split inputs into

  -p, --no-progress         disable showing progress by printing procentages
  -o, --output=FORMAT       output formats default is html. values: raw, html

EXAMPLE: Compare matchers
  Most basic usage
    ./compare_matchers matchers.scm > output.html

EXAMPLE: Check matchers on string permutatinos
  Test inputs consist of the patterns {ab}^3 and the texts {abc}^4.
    ./compare_matchers --patabc=ab --patlength=3 \\
                       --txtabc=abc --txtlength=4 \\
                       matchers.scm > output.html

EXAMPLE: Check matchers on inputs from a file
  Test inputs consist of the patterns {ab}^3 and the texts {abc}^4.
    ./compare_matchers --input-file=inputs.txt \\
                       matchers.scm > output.html

EXAMPLE: Run in parallel
  Use up to 4 cpus to process the inputs (uses GNU parallel)
    seq 4 \\
      | parallel -j+0 './compare_matchers matchers.scm \\
                             --no-progress --output=raw \\
                             --chunks=4 --chunk={} > temp_raw_{}.txt' \\
      && ./compare_matchers --raw-input temp_raw_*.txt > output.html \\
      && rm temp_raw_*.txt
                            
"))

    
(main (command-line))
